@@ -5,16 +5,15 @@ warnings.warn = warn
 import os
 import zipfile
 from os.path import join
-from urllib.error import HTTPError
-from sklearn.model_selection import StratifiedKFold
-
 import pandas as pd
+import scipy
 
 from quapy.data.base import Dataset, LabelledCollection
 from quapy.data.preprocessing import text2tfidf, reduce_columns
 from quapy.data.reader import *
 from quapy.util import download_file_if_not_exists, download_file, get_quapy_home, pickled_resource
 
+
 REVIEWS_SENTIMENT_DATASETS = ['hp', 'kindle', 'imdb']
 TWITTER_SENTIMENT_DATASETS_TEST = ['gasp', 'hcr', 'omd', 'sanders',
                               'semeval13', 'semeval14', 'semeval15', 'semeval16',
@@ -46,21 +45,39 @@ UCI_DATASETS = ['acute.a', 'acute.b',
                 'wine-q-red', 'wine-q-white',
                 'yeast']
 
+LEQUA2022_TASKS = ['T1A', 'T1B', 'T2A', 'T2B']
+
+_TXA_SAMPLE_SIZE = 250
+_TXB_SAMPLE_SIZE = 1000
+
+LEQUA2022_SAMPLE_SIZE = {
+    'TXA': _TXA_SAMPLE_SIZE,
+    'TXB': _TXB_SAMPLE_SIZE,
+    'T1A': _TXA_SAMPLE_SIZE,
+    'T1B': _TXB_SAMPLE_SIZE,
+    'T2A': _TXA_SAMPLE_SIZE,
+    'T2B': _TXB_SAMPLE_SIZE,
+    'binary': _TXA_SAMPLE_SIZE,
+    'multiclass': _TXB_SAMPLE_SIZE
+}
+
 
 def fetch_reviews(dataset_name, tfidf=False, min_df=None, data_home=None, pickle=False) -> Dataset:
     """
-    Load a Reviews dataset as a Dataset instance, as used in:
-    Esuli, A., Moreo, A., and Sebastiani, F. "A recurrent neural network for sentiment quantification."
-    Proceedings of the 27th ACM International Conference on Information and Knowledge Management. 2018.
+    Loads a Reviews dataset as a Dataset instance, as used in
+    `Esuli, A., Moreo, A., and Sebastiani, F. "A recurrent neural network for sentiment quantification."
+    Proceedings of the 27th ACM International Conference on Information and Knowledge Management. 2018. <https://dl.acm.org/doi/abs/10.1145/3269206.3269287>`_.
+    The list of valid dataset names can be accessed in `quapy.data.datasets.REVIEWS_SENTIMENT_DATASETS`
+
     :param dataset_name: the name of the dataset: valid ones are 'hp', 'kindle', 'imdb'
     :param tfidf: set to True to transform the raw documents into tfidf weighted matrices
     :param min_df: minimun number of documents that should contain a term in order for the term to be
-    kept (ignored if tfidf==False)
+        kept (ignored if tfidf==False)
     :param data_home: specify the quapy home directory where collections will be dumped (leave empty to use the default
-    ~/quay_data/ directory)
+        ~/quay_data/ directory)
     :param pickle: set to True to pickle the Dataset object the first time it is generated, in order to allow for
-    faster subsequent invokations
-    :return: a Dataset instance
+        faster subsequent invokations
+    :return: a :class:`quapy.data.base.Dataset` instance
     """
     assert dataset_name in REVIEWS_SENTIMENT_DATASETS, \
         f'Name {dataset_name} does not match any known dataset for sentiment reviews. ' \
@@ -93,22 +110,25 @@ def fetch_reviews(dataset_name, tfidf=False, min_df=None, data_home=None, pickle
 
 def fetch_twitter(dataset_name, for_model_selection=False, min_df=None, data_home=None, pickle=False) -> Dataset:
     """
-    Load a Twitter dataset as a Dataset instance, as used in:
-    Gao, W., Sebastiani, F.: From classification to quantification in tweet sentiment analysis.
-    Social Network Analysis and Mining6(19), 1–22 (2016)
-    The datasets 'semeval13', 'semeval14', 'semeval15' share the same training set.
+    Loads a Twitter dataset as a :class:`quapy.data.base.Dataset` instance, as used in:
+    `Gao, W., Sebastiani, F.: From classification to quantification in tweet sentiment analysis.
+    Social Network Analysis and Mining6(19), 1–22 (2016) <https://link.springer.com/content/pdf/10.1007/s13278-016-0327-z.pdf>`_
+    Note that the datasets 'semeval13', 'semeval14', 'semeval15' share the same training set.
+    The list of valid dataset names corresponding to training sets can be accessed in
+    `quapy.data.datasets.TWITTER_SENTIMENT_DATASETS_TRAIN`, while the test sets can be accessed in
+    `quapy.data.datasets.TWITTER_SENTIMENT_DATASETS_TEST`
 
     :param dataset_name: the name of the dataset: valid ones are 'gasp', 'hcr', 'omd', 'sanders', 'semeval13',
-    'semeval14', 'semeval15', 'semeval16', 'sst', 'wa', 'wb'
+        'semeval14', 'semeval15', 'semeval16', 'sst', 'wa', 'wb'
     :param for_model_selection: if True, then returns the train split as the training set and the devel split
-    as the test set; if False, then returns the train+devel split as the training set and the test set as the
-    test set
+        as the test set; if False, then returns the train+devel split as the training set and the test set as the
+        test set
     :param min_df: minimun number of documents that should contain a term in order for the term to be kept
     :param data_home: specify the quapy home directory where collections will be dumped (leave empty to use the default
-    ~/quay_data/ directory)
+        ~/quay_data/ directory)
     :param pickle: set to True to pickle the Dataset object the first time it is generated, in order to allow for
-    faster subsequent invokations
-    :return: a Dataset instance
+        faster subsequent invokations
+    :return: a :class:`quapy.data.base.Dataset` instance
     """
     assert dataset_name in TWITTER_SENTIMENT_DATASETS_TRAIN + TWITTER_SENTIMENT_DATASETS_TEST, \
         f'Name {dataset_name} does not match any known dataset for sentiment twitter. ' \
@@ -163,11 +183,58 @@ def fetch_twitter(dataset_name, for_model_selection=False, min_df=None, data_hom
 
 
 def fetch_UCIDataset(dataset_name, data_home=None, test_split=0.3, verbose=False) -> Dataset:
+    """
+    Loads a UCI dataset as an instance of :class:`quapy.data.base.Dataset`, as used in
+    `Pérez-Gállego, P., Quevedo, J. R., & del Coz, J. J. (2017).
+    Using ensembles for problems with characterizable changes in data distribution: A case study on quantification.
+    Information Fusion, 34, 87-100. <https://www.sciencedirect.com/science/article/pii/S1566253516300628>`_
+    and
+    `Pérez-Gállego, P., Castano, A., Quevedo, J. R., & del Coz, J. J. (2019).
+    Dynamic ensemble selection for quantification tasks.
+    Information Fusion, 45, 1-15. <https://www.sciencedirect.com/science/article/pii/S1566253517303652>`_.
+    The datasets do not come with a predefined train-test split (see :meth:`fetch_UCILabelledCollection` for further
+    information on how to use these collections), and so a train-test split is generated at desired proportion.
+    The list of valid dataset names can be accessed in `quapy.data.datasets.UCI_DATASETS`
+
+    :param dataset_name: a dataset name
+    :param data_home: specify the quapy home directory where collections will be dumped (leave empty to use the default
+        ~/quay_data/ directory)
+    :param test_split: proportion of documents to be included in the test set. The rest conforms the training set
+    :param verbose: set to True (default is False) to get information (from the UCI ML repository) about the datasets
+    :return: a :class:`quapy.data.base.Dataset` instance
+    """
     data = fetch_UCILabelledCollection(dataset_name, data_home, verbose)
     return Dataset(*data.split_stratified(1 - test_split, random_state=0))
 
 
 def fetch_UCILabelledCollection(dataset_name, data_home=None, verbose=False) -> Dataset:
+    """
+    Loads a UCI collection as an instance of :class:`quapy.data.base.LabelledCollection`, as used in
+    `Pérez-Gállego, P., Quevedo, J. R., & del Coz, J. J. (2017).
+    Using ensembles for problems with characterizable changes in data distribution: A case study on quantification.
+    Information Fusion, 34, 87-100. <https://www.sciencedirect.com/science/article/pii/S1566253516300628>`_
+    and
+    `Pérez-Gállego, P., Castano, A., Quevedo, J. R., & del Coz, J. J. (2019).
+    Dynamic ensemble selection for quantification tasks.
+    Information Fusion, 45, 1-15. <https://www.sciencedirect.com/science/article/pii/S1566253517303652>`_.
+    The datasets do not come with a predefined train-test split, and so Pérez-Gállego et al. adopted a 5FCVx2 evaluation
+    protocol, meaning that each collection was used to generate two rounds (hence the x2) of 5 fold cross validation.
+    This can be reproduced by using :meth:`quapy.data.base.Dataset.kFCV`, e.g.:
+
+    >>> import quapy as qp
+    >>> collection = qp.datasets.fetch_UCILabelledCollection("yeast")
+    >>> for data in qp.data.Dataset.kFCV(collection, nfolds=5, nrepeats=2):
+    >>>     ...
+
+    The list of valid dataset names can be accessed in `quapy.data.datasets.UCI_DATASETS`
+
+    :param dataset_name: a dataset name
+    :param data_home: specify the quapy home directory where collections will be dumped (leave empty to use the default
+        ~/quay_data/ directory)
+    :param test_split: proportion of documents to be included in the test set. The rest conforms the training set
+    :param verbose: set to True (default is False) to get information (from the UCI ML repository) about the datasets
+    :return: a :class:`quapy.data.base.Dataset` instance
+    """
 
     assert dataset_name in UCI_DATASETS, \
         f'Name {dataset_name} does not match any known dataset from the UCI Machine Learning datasets repository. ' \
@@ -302,7 +369,7 @@ def fetch_UCILabelledCollection(dataset_name, data_home=None, verbose=False) ->
         df = pd.read_csv(data_path, header=None, encoding='utf-16', sep='\t')
 
         df[0] = df[0].apply(lambda x: float(x.replace(',', '.'))).astype(float, copy=False)
-        [df_replace(df, col) for col in range(1, 6)]
+        [_df_replace(df, col) for col in range(1, 6)]
         X = df.loc[:, 0:5].values
         if dataset_name == 'acute.a':
             y = binarize(df[6], pos_class='yes')
@@ -482,5 +549,78 @@ def fetch_UCILabelledCollection(dataset_name, data_home=None, verbose=False) ->
     return data
 
 
-def df_replace(df, col, repl={'yes': 1, 'no':0}, astype=float):
-    df[col] = df[col].apply(lambda x:repl[x]).astype(astype, copy=False)
\ No newline at end of file
+def _df_replace(df, col, repl={'yes': 1, 'no':0}, astype=float):
+    df[col] = df[col].apply(lambda x:repl[x]).astype(astype, copy=False)
+
+
+def fetch_lequa2022(task, data_home=None):
+    """
+    Loads the official datasets provided for the `LeQua <https://lequa2022.github.io/index>`_ competition.
+    In brief, there are 4 tasks (T1A, T1B, T2A, T2B) having to do with text quantification
+    problems. Tasks T1A and T1B provide documents in vector form, while T2A and T2B provide raw documents instead.
+    Tasks T1A and T2A are binary sentiment quantification problems, while T2A and T2B are multiclass quantification
+    problems consisting of estimating the class prevalence values of 28 different merchandise products.
+    We refer to the `Esuli, A., Moreo, A., Sebastiani, F., & Sperduti, G. (2022).
+    A Detailed Overview of LeQua@ CLEF 2022: Learning to Quantify.
+    <https://ceur-ws.org/Vol-3180/paper-146.pdf>`_ for a detailed description
+    on the tasks and datasets.
+
+    The datasets are downloaded only once, and stored for fast reuse.
+
+    See `lequa2022_experiments.py` provided in the example folder, that can serve as a guide on how to use these
+    datasets.
+
+
+    :param task: a string representing the task name; valid ones are T1A, T1B, T2A, and T2B
+    :param data_home: specify the quapy home directory where collections will be dumped (leave empty to use the default
+        ~/quay_data/ directory)
+    :return: a tuple `(train, val_gen, test_gen)` where `train` is an instance of
+        :class:`quapy.data.base.LabelledCollection`, `val_gen` and `test_gen` are instances of
+        :class:`quapy.protocol.SamplesFromDir`, i.e., are sampling protocols that return a series of samples
+        labelled by prevalence.
+    """
+
+    from quapy.data._lequa2022 import load_raw_documents, load_vector_documents, SamplesFromDir
+
+    assert task in LEQUA2022_TASKS, \
+        f'Unknown task {task}. Valid ones are {LEQUA2022_TASKS}'
+    if data_home is None:
+        data_home = get_quapy_home()
+
+    URL_TRAINDEV=f'https://zenodo.org/record/6546188/files/{task}.train_dev.zip'
+    URL_TEST=f'https://zenodo.org/record/6546188/files/{task}.test.zip'
+    URL_TEST_PREV=f'https://zenodo.org/record/6546188/files/{task}.test_prevalences.zip'
+
+    lequa_dir = join(data_home, 'lequa2022')
+    os.makedirs(lequa_dir, exist_ok=True)
+
+    def download_unzip_and_remove(unzipped_path, url):
+        tmp_path = join(lequa_dir, task + '_tmp.zip')
+        download_file_if_not_exists(url, tmp_path)
+        with zipfile.ZipFile(tmp_path) as file:
+            file.extractall(unzipped_path)
+        os.remove(tmp_path)
+
+    if not os.path.exists(join(lequa_dir, task)):
+        download_unzip_and_remove(lequa_dir, URL_TRAINDEV)
+        download_unzip_and_remove(lequa_dir, URL_TEST)
+        download_unzip_and_remove(lequa_dir, URL_TEST_PREV)
+
+    if task in ['T1A', 'T1B']:
+        load_fn = load_vector_documents
+    elif task in ['T2A', 'T2B']:
+        load_fn = load_raw_documents
+
+    tr_path = join(lequa_dir, task, 'public', 'training_data.txt')
+    train = LabelledCollection.load(tr_path, loader_func=load_fn)
+
+    val_samples_path = join(lequa_dir, task, 'public', 'dev_samples')
+    val_true_prev_path = join(lequa_dir, task, 'public', 'dev_prevalences.txt')
+    val_gen = SamplesFromDir(val_samples_path, val_true_prev_path, load_fn=load_fn)
+
+    test_samples_path = join(lequa_dir, task, 'public', 'test_samples')
+    test_true_prev_path = join(lequa_dir, task, 'public', 'test_prevalences.txt')
+    test_gen = SamplesFromDir(test_samples_path, test_true_prev_path, load_fn=load_fn)
+
+    return train, val_gen, test_gen
+
